#include "8puzzle.h"
#include <iostream>
#include <cstdio>
using namespace std;

//IMPORTANT: The numbers 1 to 8 are actual puzzle pieces.
//			 The number 0 is the blank piece
//			 The number 10 is used in goal states to indicate Dont-Care.


//int eightState::getManhattanH(){
//	int i,j,l=0;
//	int hval=0;
//	for (i=0;i<2;i++){//find manhattan distance for first 2 rows.
//		for (j=0;j<3;j++){
//			l++;
//			hval+=(abs(xPos[l]-i)+abs(yPos[l]-j));
//		}
//	}
//	for (j=0;j<2;j++){//then find distance for first 2 cols of last row.
//		l++;
//		hval+=(abs(xPos[l]-i)+abs(yPos[l]-j));
//	}
//
//	//finally find distance of 0 from last row,last col
//	//hval+=(4-xPos[0]-yPos[0]);
//	return hval;
//}


//the num-out-of-place heuristic
int eightState::getNumOutOfPlace(){
	Byte * p = &(board[0][0]);
	int retVal=0;
	for (int i=0;i<8;i++)
		if (p[i]!=i+1)
			retVal++;
	return retVal;
}

//set a particular piece of the eight-puzzle board.
void eightState::setCell(int x, int y, int value){
	board[x][y]=value;
	xPos[value]=x;
	yPos[value]=y;
}

//look-up-table-heuristic
int eightState::getLutH(){
	int h1,h2,h3,h4;
	//gets the heuristic for current positions of (1,2,0), (3,4,0), (5,6,0) and (7,8,0)
	//These heuristics are the actual number of steps taken to solve the partial boards.

	h1=lut[0][xPos[1]+3*yPos[1]][xPos[2]+3*yPos[2]][xPos[0]+3*yPos[0]];
	h2=lut[1][xPos[3]+3*yPos[3]][xPos[4]+3*yPos[4]][xPos[0]+3*yPos[0]];
	h3=lut[2][xPos[5]+3*yPos[5]][xPos[6]+3*yPos[6]][xPos[0]+3*yPos[0]];
	h4=lut[3][xPos[7]+3*yPos[7]][xPos[8]+3*yPos[8]][xPos[0]+3*yPos[0]];

	//return the max of the heuristics
	return max(max(h1,h2),max(h3,h4));
}

int eightState::getBubbleH(){
	int tmp1[9],tmp2[9], i,j;
	int ilate1[9]={10,1,2,3,4,5,6,7,8};
	int ilate2[9]={10,1,4,7,2,5,8,3,6};

	for (i=0;i<3;i++){//find manhattan distance
		for (j=0;j<3;j++){
			tmp1[j+3*i]=ilate1[board[i][j]];
			tmp2[i+3*j]=ilate2[board[i][j]];
		}
	}
	int swaps1=0,swaps2=0,temp2;
	for (i=0;i<8;i++){
		for (j=i+1;j<9;j++){
			if (tmp1[i]>tmp1[j]){
				swaps1++;
				temp2=tmp1[i];
				tmp1[i]=tmp1[j];
				tmp1[j]=temp2;
			}
		}
	}
	for (i=0;i<8;i++){
		for (j=i+1;j<9;j++){
			if (tmp2[i]>tmp2[j]){
				swaps2++;
				temp2=tmp2[i];
				tmp2[i]=tmp2[j];
				tmp2[j]=temp2;
			}
		}
	}
	return swaps1;//(swaps1>swaps2? swaps2: swaps1);
}

//generalised manhattan distance to a given goal state
int eightState::getManhattanH(){
	int i,j,g;
	int hval=0;
	for (i=0;i<3;i++){//find manhattan distance
		for (j=0;j<3;j++){
			g=goal->board[i][j];
			if (g!=0 && g!=10){//0 is the blank and '10' is a Dont-Care piece
				hval+= (abs(xPos[g]-i) + abs(yPos[g]-j));
			}
		}
	}
	return hval;
}

//manhattan distance+penalty
int eightState::getManhattanPenaltyH(){
	int i,j,g,dup,dist;
	int hval=0;
	for (i=0;i<3;i++){//find manhattan distance
		for (j=0;j<3;j++){
			g=goal->board[i][j];
			if (g!=0 && g!=10){
				dist= abs(xPos[g]-i) + abs(yPos[g]-j);
				hval+= dist;

				//penalty heuristic
				//if 2 pieces are in each other's places, and
				//they are adjacent. like 213... (goal is 123...)
				//then, add one penalty point
				if (dist==1){
					dup=goal->board[xPos[g]][yPos[g]];
					if ( dup!=0 && dup!=10 && xPos[dup]==i && yPos[dup]==j)
						hval++;
				}
			}
		}
	}
	return hval;
}

//manhattan distance+penalty (OLD implementation, NON Monotone, Closed List Updates)
int eightState::getManhattanPenaltyOldH(){
	int i,j,g,dup,dist;
	int hval=0;
	for (i=0;i<3;i++){//find manhattan distance
		for (j=0;j<3;j++){
			g=goal->board[i][j];
			if (g!=0 && g!=10){
				dist= abs(xPos[g]-i) + abs(yPos[g]-j);
				hval+= dist;

				//penalty heuristic
				//if 2 pieces are in each other's places, and
				//they are adjacent. like 213... (goal is 123...)
				//then, add one penalty point
				if (dist>0){
					dup=goal->board[xPos[g]][yPos[g]];
					if ( dup!=0 && dup!=10 && xPos[dup]==i && yPos[dup]==j)
						hval++;
				}
			}
		}
	}
	return hval;
}

//compare with current goal state.
bool eightState::isGoalState(){
	static int i,j;
	for (i=0;i<3;i++)
		for (j=0;j<3;j++)
			if (goal->board[i][j]!=10 && goal->board[i][j]!=board[i][j])
				return false;

	return true;
}

//set the heuristic to be used. An integer flag(heuristic) is set here.
void eightState::setHeuristic(char* h){
	if (strcmp(h,"lut")==0){
		heuristic=4;
	}else if (strcmp(h,"manhattan")==0){
		heuristic=2;
	}else if (strcmp(h,"penalty")==0){
		heuristic=3;
	}else if (strcmp(h,"bfs")==0){
		heuristic=0;
	}else if (strcmp(h,"outofplace")==0){
		heuristic=1;
	}else if (strcmp(h,"penalty+lut")==0){
		heuristic=5;
	}else if (strcmp(h,"2*manhattan")==0){
		heuristic=6;
	}else if (strcmp(h,"penalty_old")==0){
		heuristic=7;
	}else if (strcmp(h,"5*penalty+lut")==0){
		heuristic=8;
	}else if (strcmp(h,"bubble")==0){
		heuristic=9;
	}
}

//return the heuristic for this state.
int eightState::getHValue(){
	switch (heuristic){
	case 0:
		return 0;
	case 1:
		return getNumOutOfPlace();
	case 2:
		return getManhattanH();
	case 3:
		return getManhattanPenaltyH();
	case 4:
		return getLutH();
	case 5:
		return max(getLutH(),getManhattanPenaltyH());
	case 6:
		return 2*getManhattanH();
	case 7:
		return getManhattanPenaltyOldH();
	case 8:
		return 5*max(getLutH(),getManhattanPenaltyH());
	case 9:
		return getBubbleH();
	default:
		return 0;
	}
}

//makes a move. here, x is the displacement in x direction.
//and y is the displacement in y direction. (+1 to -1 each)
bool eightState::makeMove(Byte x, Byte y){
	Byte x0=xPos[0]+x;
	Byte y0=yPos[0]+y;
	if (x0<0 || x0>2 || y0<0 || y0>2)
		return false;
	Byte value=board[x0][y0];

	board[x0][y0]=0;
	board[xPos[0]][yPos[0]]=value;

	xPos[value]=xPos[0];
	yPos[value]=yPos[0];

	xPos[0]=x0;
	yPos[0]=y0;
	return true;
}

//make the successors
list<worldState *>& eightState::generateSuccessors(){
	eightState tmpState;
	eightState* rPtr;

	freeMemory();//delete any existing data

	tmpState.setGValue(gValue+1);
	tmpState.parent=this;
	tmpState.setState(board);

	for (Byte dx=-1;dx<=+1;dx+=2){//simulate the 4 possible moves
		if (tmpState.makeMove(dx,0)){//if the move is valid
			if (!parent || !tmpState.isEqual(*parent)){
				rPtr=new eightState;
				(*rPtr)=tmpState;	//make a new state node,and add to list.
				potentialSuccessors.push_front((worldState*)rPtr);
			}
			tmpState.setState(board);//reset the board.
		}
	}
	for (Byte dy = -1 ; dy <= +1; dy += 2 ){
		if (tmpState.makeMove(0,dy)){//if the move is valid
			if (!parent || !tmpState.isEqual(*parent)){
				rPtr=new eightState;
				(*rPtr)=tmpState;	//make a new state node,and add to list.
				potentialSuccessors.push_front((worldState*)rPtr);
			}
			tmpState.setState(board);//reset the board.
		}
	}

	return potentialSuccessors;
}

//print the state
void eightState::printState(){
	for (Byte i=0;i<3;i++){
		for (Byte j=0;j<3;j++)
			cout<<(int)board[i][j]<< " ";
		cout<<endl;
	}
	cout<<"f="<<getFValue()<<", g="<<getGValue()<<", h="<<getHValue();
	cout<<"\n~~~~~~~~~";
}

//return true if states are equal.
bool eightState::isEqual(worldState& w){
	static int i,j;
	eightState* w2=(eightState*)&w;
	for (i=0;i<3;i++)
		for (j=0;j<3;j++)
			if (w2->board[i][j]!=board[i][j])
				return false;

	return true;
}

//return true or false depending on which configuration the board is in.
//There are 2 sets of states, which cant be reached from each other.
//If the board is in one of these sets, function returns true, the other returns false.
//Basically, if 2 boards have different "configuration", they cant be reached from one another.
bool eightState::getConfiguration(){
	int tileAt[10]={0,0,0,0,0,0,0,0,0,0};//stores the tile values of 1...9 tiles
					//in zig-zag order.
	int netValue=0;//for each tile, sum up the number of tiles that
					//come after it, and are smaller than it.
	int cursor=1;//current position in board zig-zag pattern.

	for (Byte i=0;i<3;i++){
		for (Byte j=0;j<3;j++){
			tileAt[cursor]=board[i][j];//get the value of the current tile

			if(tileAt[cursor]!=0 && tileAt[cursor]!=10){
				for (Byte k=cursor-1;k>0;k--)//for all numbers before this, check if this tile has smaller value.
					if ((int)tileAt[k]!=10 && (int)tileAt[k]>(int)tileAt[cursor])
						netValue++;
			}
			cursor++;
		}
	}

	//if the net value is even,then board belongs to one set, else the other set.
	return (netValue%2==1);
}

//sets the goal state
void eightState::setGoalState(eightState *s){
	goal=s;
}

//compare board configurations
bool eightState::isSolvable(){
	return (this->getConfiguration()==goal->getConfiguration());
}

//set the board using an array
void eightState::setState(Byte b[3][3]){
	for (Byte i=0;i<3;i++)
		for (Byte j=0;j<3;j++){
			board[i][j]=b[i][j];
			if (board[i][j]==10)
				continue;
			xPos[b[i][j]]=i;
			yPos[b[i][j]]=j;
		}
}

//Save the lookup table.
void eightState::saveLUT(){
	FILE* fp = fopen("data.txt","w");
	for (int w=0;w<4;w++)
		for (int i=0;i<9;i++)
			for (int j=0;j<9;j++)
				for (int k=0;k<9;k++)
					fprintf(fp,"%d ",lut[w][i][j][k]);
	fclose(fp);
}

//Load the lookup table
void eightState::loadLUT(){
	FILE* fp = fopen("data.txt","r");
	for (int w=0;w<4;w++)
		for (int i=0;i<9;i++)
			for (int j=0;j<9;j++)
				for (int k=0;k<9;k++)
					fscanf(fp,"%d",&lut[w][i][j][k]);

	fclose(fp);
}

//Generate the look up table.
//This method creates many, many boards, and solves them,
//and remembers how many steps are needed to solve them, by
//storing them in a big array.
void eightState::generateLUT(){
	int a120[]={1,2,0}, b120[]={3,4,5,6,7,8};
	int a340[]={3,4,0}, b340[]={1,2,5,6,7,8};
	int a560[]={5,6,0}, b560[]={1,2,3,4,7,8};
	int a780[]={7,8,0}, b780[]={1,2,3,4,5,6};

	aStar eightProblem;
	eightState tmp,goal[4];
	worldState * result;

	Byte gInt1[3][3]={{1,2,10},{10,10,10},{10,10,10}};
	Byte gInt2[3][3]={{10,10,3},{4,10,10},{10,10,10}};
	Byte gInt3[3][3]={{10,10,10},{10,5,6},{10,10,10}};
	Byte gInt4[3][3]={{10,10,10},{10,10,10},{7,8,10}};

	goal[0].setState(gInt1);
	goal[1].setState(gInt2);
	goal[2].setState(gInt3);
	goal[3].setState(gInt4);

	for (int w=0;w<4;w++){
		eightState::setGoalState(&goal[w]);
		for (int i=0;i<9;i++){
			for (int j=0;j<9;j++){
				for (int k=0;k<9;k++){
					if (i!=j && j!=k && i!=k){
						cout<<"("<<w<<i<<j<<k<<")";
						if (w==0)
							tmp.generateBoard(a120,b120,i,j,k);
						else if (w==1)
							tmp.generateBoard(a340,b340,i,j,k);
						else if (w==2)
							tmp.generateBoard(a560,b560,i,j,k);
						else
							tmp.generateBoard(a780,b780,i,j,k);

						//we assume this problem can be solved
						result=eightProblem.runAStar(&tmp);

						if (!result){
							cout<<"Cannot reach goal node\n";
						}else{
							lut[w][i][j][k]=eightProblem.getStats().lengthOfOptimalPath;
							cout<<"Optimal Path "<<(int)lut[w][i][j][k]<<" found\n";
						}
						eightProblem.cleanUp();
					}
				}
			}
		}
	}
	saveLUT();
}

//generate a board where only 3 numbers(arr1) are in specific places (a,b and c),
//and remaining (arr2) are in the spaces between (their position dont matter)
void eightState::generateBoard(int arr1[],int arr2[],int a,int b, int c){
	int cnt=0,pos;
	int i,j;
	for (j=0;j<3;j++)
	for (i=0;i<3;i++)
	{
		pos=i+3*j;
		if (pos==a)
			setCell(i,j,arr1[0]);
		else if (pos==b)
			setCell(i,j,arr1[1]);
		else if (pos==c)
			setCell(i,j,arr1[2]);
		else
			setCell(i,j,arr2[cnt++]);
	}
}

//static variables are initialised here:

eightState* eightState::goal=0;
Byte eightState::lut[4][9][9][9]={0};
int eightState::heuristic=0;
