#include "aStar.h"
#include <iostream>
#include <time.h>
using namespace std;

int max(int a , int b){
	return (a>b?a:b);
}

int min(int a , int b){
	return (a<b?a:b);
}

//f(n)=g(n)+h(n), where g(n) is distance of n from Start node,
//and h(n) is the heuristic distance estimate to Goal node.
worldState::worldState(){
	gValue=0;
	parent=NULL;
}

int worldState::getGValue(){
	return gValue;
}

int worldState::getFValue(){
	return getGValue()+getHValue();
}

worldState * worldState::getParent(){
	return parent;
}

bool worldState::setParent(worldState * p){
	if (parent!=NULL){
		if (parent->getGValue()<=p->getGValue())
			return false;//there is already a better parent
	}
	parent=p;
	return true;
}

void worldState::addChild(worldState * p){
	children.push_front(p);
}

void worldState::setGValue(int newG){//force set the g value without propogation
	gValue=newG;
}

void worldState::propogateGValue(int newG){//propogate the g value through the system
	if (gValue<=newG)
		return;
	else{
		gValue=newG;//set new g value

		//and propogate the new g value through the children
		list<worldState*>::iterator i;
		for (i=children.begin();i!=children.end();i++){
			if ((*i)->parent==this){//the parent of this child is still "this" node
				(*i)->propogateGValue(gValue+1);
			}else{//the parent has changed
				if ((*i)->setParent(this)){//if "this" is a better parent
					(*i)->propogateGValue(gValue+1);
				}
			}
		}
	}
}

void worldState::freeMemory(){
	if (potentialSuccessors.size()==0)
		return;
	list<worldState*>::iterator i;
	for (i=potentialSuccessors.begin();i!=potentialSuccessors.end();i++){
		delete (*i);
	}
	children.clear();
	potentialSuccessors.clear();
}

worldState::~worldState(){
}

bool worldState::isMonotone(){
	if (parent)
		if(parent->getHValue()>getHValue()+1)
			return false;

	return true;
}

void worldState::doWhenAddToClosedList(){
}

/////////////////////////////////////////////////////////////////////////////////////

aStarStats::aStarStats(){
	numNodesExpanded=0;
	numNodesGenerated=0;
	closedListLength=0;
	openListLength=0;
	numClosedListRevisits=0;
	numOpenListRevisits=0;
	costOfBestNodeSearch=0;
	closedListUpdates=0;
	lengthOfOptimalPath=0;
	monotonous=true;
}

void aStarStats::print(){
	if (monotonous)
		cout<<"Monotonous"<<endl;
	else
		cout<<"Non-monotonous"<<endl;

	cout<<"Length of Optimal Path       : "<<lengthOfOptimalPath<<endl;
	cout<<"Number of Nodes expanded     : "<<numNodesExpanded<<endl;
	cout<<"Closed List nodes Updated    : "<<closedListUpdates<<endl;
	//cout<<"Number of Nodes generated    : "<<numNodesGenerated<<endl;
	//cout<<"Final Length of Open List    : "<<openListLength<<endl;
	//cout<<"Final Length of Closed List  : "<<closedListLength<<endl;
	//cout<<"Open List nodes Revisited    : "<<numOpenListRevisits<<endl;
	//cout<<"Closed List nodes Revisited  : "<<numClosedListRevisits<<endl;
	//cout<<"Cost of search in Open List  : "<<costOfBestNodeSearch<<endl;
	cout<<"Time taken (seconds)         : "<<timeTaken<<endl;
}

/////////////////////////////////////////////////////////////////////////////////////

void aStar::cleanUp(){
	list<worldState*>::iterator i;
	for(i=openList.begin();i!=openList.end();i++){
		(*i)->freeMemory();
	}
	for(i=closedList.begin();i!=closedList.end();i++){
		(*i)->freeMemory();
	}
	closedList.clear();
	openList.clear();
}

aStarStats aStar::getStats(){
	return statistics;
}

//returns the best node from the open list.
worldState* aStar::popBestNode(){
	worldState * p=NULL;

	statistics.numNodesExpanded++;

	//search through the list for best node
	list<worldState*>::iterator i,j;
	for (i=openList.begin();i!=openList.end();i++){

		statistics.costOfBestNodeSearch++;

		if (p==NULL || (*i)->getFValue()<p->getFValue()){
			p=(*i);
			j=i;
		}
	}

	//remove the element from openlist
	openList.erase(j);//erase doesnt throw an exception. if j not exist, it is ignored.

	return p;//if it is still null, then it means algo ended.
}

worldState * aStar::addToClosedList(worldState *p){//this method returns true if p is the goal state
	if (!p->isMonotone())
		statistics.monotonous=false;

	p->doWhenAddToClosedList();

	if (p->isGoalState()){
		return p;
	}else{
		closedList.push_front(p);
	}
	return NULL;
}

bool aStar::addToOpenList(worldState *p){//returns false if the node can be deleted.

	list<worldState*>::iterator i;

	for(i=openList.begin();i!=openList.end();i++){
		if ( (*i)->isEqual(*p) ){
			//found the node on open list!
			statistics.numOpenListRevisits++;

			p->getParent()->addChild((*i));//add existing node to children list of p's parent

			if ((*i)->getGValue()>p->getGValue()){//if new path is cheaper
				(*i)->setParent(p->getParent());
			}
			return false;
		}
	}

	for(i=closedList.begin();i!=closedList.end();i++){
		if ((*i)->isEqual(*p)){
			//found the node on closed list!
			statistics.numClosedListRevisits++;

			if ((*i)->getGValue()>p->getGValue()){//if new path is cheaper
				statistics.closedListUpdates++;
				(*i)->setParent(p->getParent());
				(*i)->propogateGValue(p->getGValue());
			}
			return false;
		}
	}

	//no equivalent nodes found
	openList.push_front(p);//add the input node to open list.
	if (p->getParent())
		p->getParent()->addChild(p);//add it to its parent's chosen children

	return true;
}

worldState * aStar::runAStar(worldState * startState){

	//statistics.timeTaken = clock();
	statistics.timeTaken = time(0);

	worldState * finalState=NULL;
	worldState * temp;

	addToOpenList(startState);

	while (openList.size()>0){

		temp=popBestNode();
		//temp->printState();
		//cout<<endl;
		finalState=addToClosedList(temp);
		if (finalState){
			//cout<<"found!";
			break;
		}

		list<worldState*> & cList= temp->generateSuccessors();
		statistics.numNodesGenerated+=cList.size();
		//cList.sort();

		for (list<worldState*>::iterator i=cList.begin();i!=cList.end();i++){
			if (!addToOpenList((*i))){
				delete (*i);//node not needed, can be deleted.
			}
		}
		cList.clear();//the successor list is no longer needed
	}

	statistics.openListLength=openList.size();
	statistics.closedListLength=closedList.size();

	//statistics.timeTaken=((clock()-statistics.timeTaken)*1000/CLOCKS_PER_SEC);
	statistics.timeTaken = time(0)-statistics.timeTaken;
	if (finalState)
		statistics.lengthOfOptimalPath=finalState->getGValue();
	return finalState;
}

