#ifndef ASTAR_H
#define ASTAR_H

#include <list>
//#include <windows.h>
using namespace std;

int max(int a , int b);
int min(int a , int b);

class worldState{
protected:
	worldState * parent;
	list<worldState *> children;
	list<worldState *> potentialSuccessors;
	int gValue;
public:
	//f(n)=g(n)+h(n), where g(n) is distance of n from Start node,
	//and h(n) is the heuristic distance estimate to Goal node.
	worldState();
	int getGValue();
	int getFValue();
	worldState * getParent();
	bool setParent(worldState * p);
	void addChild(worldState * p);
	void setGValue(int newG);
	void propogateGValue(int newG);
	void freeMemory();
	bool isMonotone();
	virtual void doWhenAddToClosedList();

	virtual bool isGoalState()=0;
	virtual int getHValue()=0;
	virtual void printState()=0;
	virtual bool isEqual(worldState& w)=0;
	virtual list<worldState *>& generateSuccessors()=0;
	virtual ~worldState();
};

class aStarStats{
public:
	long numNodesGenerated;
	long numNodesExpanded;
	long closedListLength;
	long openListLength;
	long numClosedListRevisits;
	long numOpenListRevisits;
	long costOfBestNodeSearch;
	long closedListUpdates;
	long lengthOfOptimalPath;
	long timeTaken;
	bool monotonous;

	aStarStats();
	void print();
};

class aStar{
protected:
	list<worldState *> openList,closedList;
	aStarStats statistics;
	//returns the best node from the open list.
	worldState* popBestNode();
	worldState* addToClosedList(worldState *p);
	bool addToOpenList(worldState *p);
public:
	aStarStats getStats();
	void cleanUp();
	worldState * runAStar(worldState * startState);
};

#endif
