#include <iostream>
#include <fstream>
using namespace std;
#include "wang.h"
#include "8puzzle.h"
#include "missionary.h"

int wang_main(int argc, char**argv);
int eight_main(int argc, char**argv);
int missionary_main(int argc, char**argv);
int generateEight();

int main(int argc, char**argv){
	if (strcmp(argv[1],"missionary")==0){
		return missionary_main(argc,argv);
	}else if (strcmp(argv[1],"eight")==0){
		return eight_main(argc,argv);
	}else if (strcmp(argv[1],"wang")==0){
		return wang_main(argc,argv);
	}else if (strcmp(argv[1],"generate-lut")==0){
		return generateEight();
	}
	return -1;
}

int wang_main(int argc, char**argv){
	aStar wangProblem;

	wangState start;
	worldState * result;
	wangTile tile;
	int outMode;

	char data[64],heuristic[64],tileid;
	int a,b,c,d,e,num,tclass;
	ifstream file1(argv[2]);
	file1>>data>>a>>b;
	wangState::setDimensions(a,b);
	file1>>data>>heuristic;//read the heuristic
	wangState::setHeuristic(heuristic);

	file1>>data>>num;//read number of tiles

	for (int i=0;i<num;i++){//read the tiles
		file1>>a>>b>>c>>d>>tileid>>e>>tclass;
		tile.initTile(a,b,c,d,tileid,tclass);
		wangState::addTile(tile);
		start.setNumTiles(tclass,e);
	}
	file1.close();

	//randomise the starting tiles
	if (strcmp(heuristic,"misplaced")==0){
		srand(time(0));
		start.fillEmptySpaces();
		int cnt=start.getBadTileCount();
		int baddies[128];
		int numbad = start.getRandomBadTiles(baddies,cnt);
		start.returnTilesToPool(baddies,numbad );
		start.fitTilesIntoPlace();
		cout<<"Starting:"<<endl;
		start.printState();
	}

	cout<<"Solving. Please Wait... Heuristic ("<<heuristic<<")"<<endl;

	result=wangProblem.runAStar(&start);
	if (!result){
		cout<<"Cannot reach goal node\n";
	}else{
		result->printState();
	}

	cout<<"\nStatistics:\n";
	wangProblem.getStats().print();

	getchar();
	exit(0);
}

int generateEight(){
	eightState::setHeuristic("manhattan");
	eightState::generateLUT();
	cout<<"\nDone\n";
	return 0;
}

int eight_main(int argc, char**argv){
	aStar eightProblem;

	Byte sInt[3][3];
	Byte gInt[3][3];

	eightState start,goal;
	worldState * result;

	start.loadLUT();

	ifstream file1(argv[2]);
	char data[64],heuristic[64],mode[64];
	int outMode=2,tmpInt;
	file1>>data;

	for (int i=0;i<3;i++)
		for (int j=0;j<3;j++){
			file1>>tmpInt;
			sInt[i][j]=tmpInt;
		}
	file1>>data;
	for (int i=0;i<3;i++)
		for (int j=0;j<3;j++){
			file1>>tmpInt;
			gInt[i][j]=tmpInt;
		}

	file1>>data>>heuristic>>data>>mode;
	file1.close();

	if (strcmp(mode,"stats")==0){
		outMode=0;
	}else if (strcmp(mode,"path")==0){
		outMode=1;
	}else if (strcmp(mode,"full")==0){
		outMode=2;
	}
	eightState::setHeuristic(heuristic);


	start.setState(sInt);
	goal.setState(gInt);

	eightState::setGoalState(&goal);

	cout<<"Solving. Please wait... (Heuristic: "<<heuristic<<")"<<endl;

	if (start.isSolvable())
		result=eightProblem.runAStar(&start);
	else
		result=false;

	if (!result){
		cout<<"Cannot reach goal node\n";
		exit(0);
	}else{
		if (outMode==1 || outMode==2){
			while (result){
				result->printState();
				cout<<endl;
				result=result->getParent();
			}
		}
		if (outMode==0 || outMode==2){
			cout<<"\nStatistics:\n";
			eightProblem.getStats().print();
		}
	}

	eightProblem.cleanUp();

	getchar();
	exit(0);
}

//For Missionary-Cannibal Problem
int missionary_main(int argc, char**argv){
	aStar cannibalProblem;
	riverState start;
	worldState * result;

	char data[64];//data is used just to skip over useless parts of the file.
	char heuristic[64];
	int maxM, maxC, boatCap;
	int startM, startC, startBoat;
	int outMode=0;

	ifstream file1(argv[2]);
		file1>>data>>maxM>>data>>maxC>>data>>boatCap;
		file1>>data>>startM>>data>>startC>>data>>startBoat;
		file1>>data>>heuristic;
		file1>>data>>data;
	file1.close();

	if (strcmp(data,"path")==0)
		outMode=0;
	else if (strcmp(data,"stats")==0)
		outMode=1;
	else if (strcmp(data,"all")==0)
		outMode=2;

	riverState::setHeuristic(heuristic);
	riverState::setLimits(maxM,maxC,boatCap);
	start.setState(startM,startC,startBoat);

	result=cannibalProblem.runAStar(&start);
	cout<<"Solving. Please wait... (Heuristic: "<<heuristic<<")"<<endl;

	if (!result){
		cout<<"Cannot reach goal node\n";
	}else{
		if (outMode==0 || outMode==2){
			while (result){
					result->printState();
					cout<<endl;
					result=result->getParent();
			}
		}
	}

	if (outMode==1 || outMode==2){
		cout<<"\nStatistics:\n";
		cannibalProblem.getStats().print();
	}

	getchar();
	exit(0);
}
