#include "missionary.h"
#include <iostream>
using namespace std;

bool riverState::outNumber(int a, int b){
	if (b==0)
		return false;
	else
		return (a>b);
}

void riverState::setHeuristic(char * h){
	if (strcmp("bfs",h)==0){
		heuristic=0;
	}else if (strcmp("basic",h)==0){
		heuristic=1;
	}else if (strcmp("extended",h)==0){
		heuristic=2;
	}else if (strcmp("bignum",h)==0){
		heuristic=3;
	}
}

bool riverState::isValidState(){
	if (numCannibals>maxCannibals || numMissionaries>maxMissionaries ||
		numCannibals<0 || numMissionaries<0)
			return false;
	if (outNumber(numCannibals,numMissionaries))
		return false;
	if (outNumber(maxCannibals-numCannibals,maxMissionaries-numMissionaries))
		return false;

	return true;
}

bool riverState::isGoalState(){
	if (numCannibals==0 && numMissionaries==0 && boatSide==1)
		return true;
	else
		return false;
}
int riverState::getHValue(){
	switch (heuristic){
	case 0://bfs
		return 0;
	case 1://basic heuristic
		return (numMissionaries+numCannibals -1)/2;
	case 2://extended heuristic
		if (boatSide==0)
			return (numMissionaries+numCannibals-boatCapacity)*2/(boatCapacity-1) + 1;
		else
			return (numMissionaries+numCannibals+1-boatCapacity)*2/(boatCapacity-1) + 2;
	case 3://a big number
		return 99999;
	}
}

list<worldState *>& riverState::generateSuccessors(){
	riverState tmpState;
	riverState *rPtr;
	int nc, nm;
	int crossC, crossM;

	freeMemory();//delete any existing data

	//the boat can take 1 to boatCapacity number of people across.
	//the number of cannibals can be 0 to numCannibals
	//the number of missionaries can be 0 to numMissionaries

	if (boatSide==0){
		nc=numCannibals;
		nm=numMissionaries;
	}else{
		nc=maxCannibals-numCannibals;
		nm=maxMissionaries-numMissionaries;
	}

	tmpState.boatSide=1-boatSide;//move boat to other side of river
	tmpState.setGValue(gValue+1);
	tmpState.parent=this;

	for (crossC=0;crossC<=min(nc,boatCapacity);crossC++){
		for (crossM=0;crossM<=min(nm,boatCapacity);crossM++){
			if (crossC+crossM>0 && crossC+crossM<=boatCapacity && !outNumber(crossC,crossM)){
				tmpState.numCannibals	=	nc-crossC;
				tmpState.numMissionaries=	nm-crossM;
				if (boatSide==1){
					tmpState.numCannibals	=	maxCannibals-tmpState.numCannibals;
					tmpState.numMissionaries=	maxMissionaries-tmpState.numMissionaries;
				}
				if (tmpState.isValidState()){
					rPtr=new riverState;
					(*rPtr)=tmpState;
					potentialSuccessors.push_front((worldState*)rPtr);
				}
			}
		}
	}

	return potentialSuccessors;
}

void riverState::printState(){
	cout<<"<"<<numMissionaries<<","<<numCannibals<<","<<boatSide<<"> => f="<<getFValue()<<"\t g="<<getGValue()<<"\t h="<<getHValue();
}

bool riverState::isEqual(worldState& w){
	riverState& w2=dynamic_cast<riverState&>(w);

	if (numMissionaries==w2.numMissionaries &&
		numCannibals==w2.numCannibals &&
		boatSide==w2.boatSide)
		return true;
	else
		return false;
}

void riverState::setState(int numC,int numM, int boatS){
	numMissionaries=numM;
	numCannibals=numC;
	boatSide=boatS;
}

void riverState::setLimits(int numC,int numM, int boatS){
	maxMissionaries=numM;
	maxCannibals=numC;
	boatCapacity=boatS;
}

int riverState::heuristic=2;
int riverState::boatCapacity=2;
int riverState::maxCannibals=3;
int riverState::maxMissionaries=3;
