#include <iostream>
#include "wang.h"
using namespace std;

/////////////////////////WANG TILE///////////////////////

wangTile::wangTile(){
	for (int i=0;i<4;i++)
		colors[i]=0;
}

void wangTile::initTile(int right, int below, int left, int above,char tid, int tClass){
	colors[0]=right;
	colors[1]=below;
	colors[2]=left;
	colors[3]=above;
	tileID=tid;
	tileClass=tClass;
}

int wangTile::getLeft(){
	return colors[2];
}
int wangTile::getRight(){
	return colors[0];
}
int wangTile::getAbove(){
	return colors[3];
}
int wangTile::getBelow(){
	return colors[1];
}

/////////////////////////WANG STATE///////////////////////

wangState::wangState(){
	xPos=yPos=0;
	for (int i=0;i<MAX_TILE_CLASSES;i++)
		numTiles[i]=0;

	for (int i=0;i<WIDTH;i++)
		for (int j=0;j<HEIGHT;j++)
			board[i][j]=0;
}

void wangState::addTile(wangTile t){
	if (numWangTiles<MAX_TILES){
		tileset[numWangTiles]=t;
		numWangTiles++;
	}
}

bool wangState::isGoalState(){
//#ifdef RANDOMISE
if (heuristic==3){
	if(getBadTileCount()<=MAX_BAD_TILES){
		return true;//if number of bad tiles less than/equal to threshold
	}else{
		return false;
	}
}
//#endif

	if (yPos>=HEIGHT)//keep filling row by row, so once we fill all elements,
		return true;//the yPosition will be outside the grid (yPos>=HEIGHT)
	else			//ie: yPos has crossed the last row to be filled.
		return false;
}

int wangState::getNumPossibleMoves(){
	int counter=0;
	for (int i=0;i<numWangTiles;i++){//try placing each tile
		if (canPlace(tileset[i],xPos,yPos))
			counter++;
	}
	return counter;
}

void wangState::setHeuristic(char * p){
	if (strcmp(p,"basic")==0)
		heuristic=0;
	else if (strcmp(p,"maxmoves")==0)
		heuristic=1;
	else if (strcmp(p,"bfs")==0)
		heuristic=2;
	else if (strcmp(p,"misplaced")==0)
		heuristic=3;
}

int wangState::getHValue(){//number of empty tile positions
	switch (heuristic)
	{
	case 0:
		return ((WIDTH*HEIGHT)-(xPos+(yPos*WIDTH)));
	case 1:
		return ((WIDTH*HEIGHT)-(xPos+(yPos*WIDTH)))-getNumPossibleMoves();
	case 2:
		return 0;
	case 3:
		return getBadTileCount();
	}
}

void wangState::setNumTiles(int tileClassIndex,int tileCount){
	numTiles[tileClassIndex]=tileCount;
}

int wangState::remainingTiles(int tileClassIndex){//get remaining number of tiles
	return numTiles[tileClassIndex];
}

bool wangState::canPlace(wangTile& w, int xpos, int ypos, int * numProblems){
	int numViolated=0;
	if (xpos>0)
		if (board[xpos-1][ypos] && board[xpos-1][ypos]->getRight()!=w.getLeft())
			numViolated++;

	if (xpos<WIDTH-1)
		if (board[xpos+1][ypos] && board[xpos+1][ypos]->getLeft()!=w.getRight())
			numViolated++;

	if (ypos>0)
		if (board[xpos][ypos-1] && board[xpos][ypos-1]->getAbove()!=w.getBelow())
			numViolated++;

	if (ypos<HEIGHT-1)
		if (board[xpos][ypos+1] && board[xpos][ypos+1]->getBelow()!=w.getAbove())
			numViolated++;

#ifdef SUPERTILE
	if (xpos==0)
		if(board[WIDTH-1][ypos] && board[WIDTH-1][ypos]->getRight()!=w.getLeft())
			numViolated++;

	if (ypos==0)
		if(board[xpos][HEIGHT-1] && board[xpos][HEIGHT-1]->getAbove()!=w.getBelow())
			numViolated++;

	if (xpos==WIDTH-1)
		if(board[0][ypos] && board[0][ypos]->getLeft()!=w.getRight())
			numViolated++;

	if (ypos==HEIGHT-1)
		if(board[xpos][0] && board[xpos][0]->getBelow()!=w.getAbove())
			numViolated++;
#endif

	if (numProblems)
		*numProblems=numViolated;

	return (numViolated==0);
}


list<worldState *>& wangState::generateSuccessors(){
//#ifdef RANDOMISE
if (heuristic==3){
	return randomSuccessors();
}else{
//#else
	wangState tmpState;
	wangState* rPtr;

	//freeMemory();//delete any existing data

	if (yPos>=HEIGHT)//no more possible moves
		return potentialSuccessors;

	tmpState=(*this);//copy the parent onto the child
	tmpState.setGValue(gValue+1);
	tmpState.parent=this;
	tmpState.xPos++;
	if (tmpState.xPos==WIDTH){
		tmpState.xPos=0;
		tmpState.yPos++;
	}

	int tClass;
	for (int i=0;i<numWangTiles;i++){//try placing each tile
		//check if there are remaining tiles, and if the tile can be placed
		tClass=tileset[i].tileClass;
		if (remainingTiles(tClass) ==0 || !canPlace(tileset[i],xPos,yPos))
			continue;

		if (remainingTiles(tClass)!=-1)//value of -1 denotes infinite tiles.
			tmpState.setNumTiles(tClass,remainingTiles(tClass)-1);
			//tmpState.setNumTiles(i,numTiles[i]-1);

		rPtr=new wangState;
		(*rPtr)=tmpState;	//make a new state node,and add to list.
		(*rPtr).board[xPos][yPos]=&tileset[i];
		potentialSuccessors.push_front((worldState*)rPtr);
	}
	return potentialSuccessors;
}
//#endif
}

void wangState::printState(){
	for (int j=0;j<HEIGHT;j++){
		for (int i=0;i<WIDTH;i++){
			if (board[i][j])
				cout<<board[i][j]->tileID<<" ";
			else
				cout<<"_ ";
		}
		cout<<endl;
	}
}

bool wangState::isEqual(worldState& w){
	wangState* tmp= (wangState*)&w;

//#ifndef RANDOMISE
if (heuristic!=3){
	if (xPos!=tmp->xPos || yPos!=tmp->yPos)
		return false;
}
//#endif

	for (int i=0;i<WIDTH;i++)
		for (int j=0;j<HEIGHT;j++)
			if (board[i][j]!=tmp->board[i][j])
				return false;

	return true;
}

void wangState::setDimensions(int w, int h){
	WIDTH=w;
	HEIGHT=h;
}

/////////////////////////////////NEW RANDOM ALGORITHM////////////////////////////////

list<worldState *>& wangState::randomSuccessors(){
	wangState* rPtr;

	int numBadTiles=getBadTileCount();

	if (numBadTiles==0)//no more bad tiles
		return potentialSuccessors;

	int maxSelections = min(RANDOM_SELECT_FACTOR,numBadTiles);
	int indices[RANDOM_SELECT_FACTOR*2];

	for (int i=0;i<RANDOM_BRANCH_FACTOR;i++){
		rPtr=new wangState;//new branch

		int selectCount = getRandomBadTiles(indices,maxSelections);

		for (int z1=0;z1<WIDTH;z1++)
		for (int z2=0;z2<HEIGHT;z2++)
			rPtr->board[z1][z2]=board[z1][z2];

		for (int z1=0;z1<MAX_TILE_CLASSES;z1++)
			rPtr->numTiles[z1]=numTiles[z1];

//		*rPtr=*this;//copy the parent onto the child
		(*rPtr).setGValue(gValue+1);
		(*rPtr).parent=this;
		(*rPtr).returnTilesToPool(indices, selectCount);
		(*rPtr).fitTilesIntoPlace();
		potentialSuccessors.push_front((worldState*)rPtr);
	}
	return potentialSuccessors;
}

void wangState::fillEmptySpaces(){
	int currentTile=0;
	for (int i=0;i<WIDTH;i++)
		for (int j=0;j<HEIGHT;j++){
			if (board[i][j]!=0)//not empty spot
				continue;
			while (remainingTiles(tileset[currentTile].tileClass)==0)//no more tiles here.
				currentTile++;//switch to next tile
			if (numWangTiles<=currentTile)//all tiles exhausted.
				return;
			board[i][j]=&tileset[currentTile];
			numTiles[tileset[currentTile].tileClass]--;
		}
}

int wangState::getBadTileCount(){//number of tiles that are non-matching
	int retval=0;
	for (int i=0;i<WIDTH;i++)
		for (int j=0;j<HEIGHT;j++)
			if (board[i][j]){
				if (!canPlace(*(board[i][j]),i,j))
					retval++;
			}

	return retval;
}

bool trueRandom(int num){
	for (int i=0;i<num;i++)
		if (rand()%2 == 0)
			return true;
	return false;
}

//this array of ints should be of length count*2, because the return value
//is of the form x1,y1,x2,y2,x3,y3...xn,yn
//the return value is number of misplaced tiles actually returned (<=count).
int wangState::getRandomBadTiles(int* indices, int count){//returns 'count' number of indices for bad tiles
	int counter=0;
	int tmpflags[MAX_WIDTH][MAX_HEIGHT]={0};
	int numprobs;

	//first random pass, to choose random misfit tiles.
	for (int i=0;i<WIDTH;i++)
		for (int j=0;j<HEIGHT;j++){
			if (count==counter)
				return counter;
			//if there is a tile here, and it doesnt fit here...
			if (board[i][j] && !canPlace(*(board[i][j]),i,j,&numprobs)){
				//50-50 chance of choosing this tile
				if (trueRandom(numprobs)){
					tmpflags[i][j]=1;//tile has been taken. dont choose in 2nd pass.
					indices[counter*2]=i;
					indices[counter*2+1]=j;
					counter++;
				}
			}
		}

	//second non-random pass if required number of tiles havent been found.
	if (counter!=count){
		for (int i=0;i<WIDTH;i++)
			for (int j=0;j<HEIGHT;j++){
				if (count==counter)
					return counter;
				//if there is a tile here, and it doesnt fit here...
				if (tmpflags[i][j]==0 && board[i][j] && !canPlace(*(board[i][j]),i,j)){
					indices[counter*2]=i;
					indices[counter*2+1]=j;
					counter++;
				}
			}
	}
	return counter;
}

void wangState::doWhenAddToClosedList(){
	static int debugger=0;
	if(heuristic!=3){
		debugger++;
		if (debugger==1000){
			printState();
			debugger=0;
		}
	}else{
		if (getBadTileCount()<=7){
			printState();
			cout<<"Misplaced: "<<getBadTileCount()<<endl;
			debugger=0;
		}
	}
}

void wangState::returnTilesToPool(int tileIndices[], int count){
	int i2;
	for (int i=0;i<count;i++){
		i2=2*i;
		//cout<<tileIndices[i2]<<","<<tileIndices[i2+1]<<endl;
		//cout<<board[tileIndices[i2]][tileIndices[i2+1]]<<endl;
		if (board[tileIndices[i2]][tileIndices[i2+1]]){
			int tmp=board[tileIndices[i2]][tileIndices[i2 +1]]->tileClass;
			numTiles[tmp]++;
			board[tileIndices[i2]][tileIndices[i2 +1]]=0;
		}
	}
}

int wangState::fitTilesIntoPlace(){
	int placer,numPlaced=0;
	for (int i=0;i<WIDTH;i++)
		for (int j=0;j<HEIGHT;j++)
			if (board[i][j]==0){
				placer=whatCanBePlaced(i,j);
				if (placer!=-1){//found something that can be placed
					board[i][j]=&tileset[placer];
					numTiles[tileset[placer].tileClass]--;
					numPlaced++;
				}
			}
	fillEmptySpaces();//fill remaining spaces with arbit tiles.
	return numPlaced;
}

//gets index of a tile that can be placed, -1 otherwise
int wangState::whatCanBePlaced(int x, int y){
	for (int i=0;i<numWangTiles;i++){
		if (remainingTiles(tileset[i].tileClass)!=0 && canPlace(tileset[i],x,y))
			return i;
	}

	return -1;
}

////////////////////////////////STATIC MEMBER INIT///////////////////////////////////

int wangState::heuristic=0;
wangTile wangState::tileset[MAX_TILES];
int wangState::numWangTiles=0;
int wangState::WIDTH=10;
int wangState::HEIGHT=10;
