#ifndef WANG_H
#define WANG_H

#include "aStar.h"

#define MAX_TILES 64
#define MAX_TILE_CLASSES 64
#define MAX_WIDTH 6
#define MAX_HEIGHT 6
#define RANDOMISE
#define SUPERTILE

#define RANDOM_SELECT_FACTOR 12 //number of tiles randomly selected for refitting
#define RANDOM_BRANCH_FACTOR 12 //number of such random branches formed.
#define MAX_BAD_TILES 0 //max allowed mismatched tiles

class wangTile{
public:
	int colors[4];
	char tileID;
	int tileClass;//the class to which this tile belongs.
				//we have a set number of tiles from each class.

	wangTile();
	void initTile(int right, int below, int left, int above, char tileID, int tileClass);
	int getLeft();
	int getRight();
	int getAbove();
	int getBelow();
};

class wangState: public worldState
{
protected:
	static int WIDTH;
	static int HEIGHT;
	static int heuristic;

	static wangTile tileset[MAX_TILES];
	static int numWangTiles;

	wangTile * board[MAX_WIDTH][MAX_HEIGHT];
	int numTiles[MAX_TILE_CLASSES];//number of tiles in each class
	int xPos, yPos;//current position in board where tile must be placed.

public:
	wangState();
	static void addTile(wangTile t);
	static void setDimensions(int w, int h);
	static void setHeuristic(char * p);

	void setNumTiles(int tileClassIndex,int tileCount);
	int remainingTiles(int tileClassIndex);//get remaining number of tiles

	bool canPlace(wangTile& w, int xpos, int ypos, int * numProblems=0);
	bool isGoalState();
	int getHValue();
	int getNumPossibleMoves();
	list<worldState *>& generateSuccessors();
	void printState();
	bool isEqual(worldState& w);


	void doWhenAddToClosedList();
	list<worldState *>& randomSuccessors();
	void fillEmptySpaces();
	int getBadTileCount();//number of tiles that are non-matching
	int getRandomBadTiles(int* indices, int count);//returns 'count' number of indices for bad tiles
	void returnTilesToPool(int tileIndices[], int count);
	int fitTilesIntoPlace();
	int whatCanBePlaced(int i, int j);//gets index of a tile that can be placed, -1 otherwise
};


#endif
