package visualise;

import javax.swing.*;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.io.BufferedReader;
import java.io.FileReader;

public class Canvas  extends JFrame {
	private static final long serialVersionUID = 1L;
	private BufferedImage canvas;
    int canvasW=512,canvasH=512;
    
    Tile tiles[][]=new Tile[6][6];
    
    int left[]=new int[24];
    int right[]=new int[24];
    int up[]=new int[24];
    int down[]=new int[24];
    String alpha[]=new String[24];

    /** Creates a new instance of Canvas */
    public Canvas() {
	super("Wang Tiles");	//set caption of window
        this.setBounds(0,0,canvasW,canvasH);
        this.setResizable(false);
               
        addWindowListener( new WindowAdapter(){
            //If this event is not handled, the application will
            //be running even after the window is closed.
            public void windowClosing(WindowEvent e){
                System.exit(0);
            }
        });        
        canvas=new BufferedImage(canvasW,canvasH,BufferedImage.TYPE_INT_RGB);
        clearImage(canvas);
        initTiles();        
        
        this.setVisible(true);
        run();
    }

    private int getIndexOfTile(String s){
    	for (int i=0;i<24;i++){
    		if (s.equals(alpha[i])){
    			return i;
    		}
    	}
    	return 0;    	
    }
    
    private void initTiles(){
    	Tile.colors=new Color[4];
    	Tile.colors[0]=new Color(255,0,0);
    	Tile.colors[1]=new Color(0,255,0);
    	Tile.colors[2]=new Color(0,0,255);
    	Tile.colors[3]=new Color(255,255,0);
        try {        	
        	BufferedReader bf=new BufferedReader(new FileReader("config.txt"));        
			for (int i=0;i<24;i++){
				String s[] = bf.readLine().split("\\s+");
				right[i]=Integer.parseInt(s[0]);
				down[i]=Integer.parseInt(s[1]);
				left[i]=Integer.parseInt(s[2]);
				up[i]=Integer.parseInt(s[3]);
				alpha[i]=s[4];
			}
			bf.close();
			
        	bf=new BufferedReader(new FileReader("tile.txt"));        
			for (int i=0;i<6;i++){
				String s[] = bf.readLine().split("\\s+");
				for (int j=0;j<6;j++){
					int index=getIndexOfTile(s[j]);
					tiles[j][i]=new Tile(right[index], down[index],left[index],up[index]);
					tiles[j][i].placeTileInBoard(20,300,32,j,i);
				}
			}
			bf.close();
			

        } catch (Exception e) {
			e.printStackTrace();
		}
    }
    
    private void clearImage(BufferedImage c){
        Graphics temp=c.getGraphics();
        temp.setColor(Color.BLACK);
        temp.fillRect(0, 0, canvasW, canvasH);
        temp.setColor(Color.WHITE);
        temp.dispose();
    }
   
//This method is used to run a thread that keeps simulating and redrawing the screen.
    public void run(){
        while(true){
            try{      
                clearImage(canvas);
                paintMethod();
                Thread.sleep(20);//paint about 50 times a second
            }
            catch (Exception e){
                System.out.println(e.toString());
            }
        }
    }
    
    void paintMethod(){
        Graphics temp1=this.getGraphics();
        Graphics temp2=canvas.getGraphics();

    	for (int i=0;i<6;i++){
    		for (int j=0;j<6;j++){
    			tiles[i][j].drawTile(temp2);
    		}
    	}

        temp2.setColor(Color.WHITE);

        //This draws the contents of the hidden canvas onto the visible JFrame		
        temp1.drawImage(canvas, this.getInsets().left, this.getInsets().top, canvasW, canvasH, this);       

        temp1.dispose();
        temp2.dispose();
    }
}
